//
// this module histograms some data for the LHC HF luminosity project.
// it is a very stupid module - you give it the LHC bin number (0-3653)
// and a "weight" and it fills the histo.  it does NOT check that the
// bin number is legitimate!
//
// Drew Baden, July 2005.
//
`timescale 1ns / 1ns


module hbunch (
	input clock,
	input nclock,		// inverted clock
	input resetn,
	input [11:0] bin,    // LHC bunch number relative to BC0
	input [5:0] weight,  // a number that will always be <48 channels
	input enable_ram_write,  // active high
	input enable_ram_read,  // active high
	input zero,          // 1 means zero bin after reading 
	output reg latch_value, 	// signals to latch "value"
	output reg [15:0] value,  // value in the bin
	output [3:0] select,
	output [1:0] hstate,
	output [9:0] addrin,
	output readit,
	output [15:0] rv1,rv2,rv3,rv4,
	output [1:0] rstate1,rstate2,rstate3,rstate4,
	output rwrite1,rwrite2,rwrite3,rwrite4 
	);
//
// declarations
//
wire latch_value_1,latch_value_2,latch_value_3,latch_value_4;
wire[15:0] mem_in_1,mem_in_2,mem_in_3,mem_in_4;
wire[15:0] ram_read_value_1,ram_read_value_2,ram_read_value_3,ram_read_value_4;
//
// ok, so here i am assuming that enable_ram_write (or enable_ram_read)
// both are asserted at the same time that data starts to flow.  so i
// will latch the 2 data inputs early relative to starting the state 
// machines to fill the rams
//
//
// we need 3564 bins in the histogram.  we will use 4 of the 1kx2B memories, so 
// the low 10 bits will be the address and we control the EN by the upper bits
// note: we will be using READ-BEFORE-WRITE mode so that all we ever do when
// histogramming is write (a nice trick)
//
reg[9:0] address_d;
assign addrin = address_d;
reg[5:0] Dvalue;
reg Dfirst, Dsecond, Dthird, Dfourth;
assign select = {Dfourth,Dthird,Dsecond,Dfirst};
//wire first  = (bin[1] == 0) && (bin[0] == 0);
//wire second = (bin[1] == 0) && (bin[0] == 1);
//wire third  = (bin[1] == 1) && (bin[0] == 0);
//wire fourth = (bin[1] == 1) && (bin[0] == 1);
reg ren;
reg ram_read;
assign readit = ram_read;
//
// here the state machine that controls everything
//
reg[11:0] rbin;
reg[5:0] rval;
reg[1:0] state;
assign hstate = state;
wire enable_state = enable_ram_read || enable_ram_write;
parameter[1:0] WAIT=0, ENABLE=1;
always @ (posedge clock)
  if (~resetn) begin
	state <= WAIT;			   
	ren <= 0;
	Dfirst <= 0;
	Dsecond <= 0;
	Dthird <= 0;
	Dfourth <= 0;
	address_d <= 0;
	Dvalue <= 0;
	rbin <= 0;
	rval <= 0;
	ram_read <= 0;
	latch_value <= 0;
	value <= 0;
  end
  else begin	  
    case (state)
		WAIT: begin
			ren <= #1 1;	
//			ram_read <= #1 0;
			rbin <= #1 bin;        
			rval <= #1 weight;
			Dfirst <= 0;
			Dsecond <= 0;
			Dthird <= 0;
			Dfourth <= 0;
			ram_read <= 0;
			latch_value <= 0;
			value <= 0;
			if (enable_state) state <= #1 ENABLE;
			else state <= #1 WAIT;
			end
		ENABLE: begin
			rbin <= #1 bin;
			rval <= #1 weight;
			address_d[9:0] <= #1 rbin[11:2];
			Dvalue <= #1 rval;
			Dfirst <= #1 (rbin[1:0] == 'h0);
			Dsecond <= #1 (rbin[1:0] == 'h1);
			Dthird <= #1 (rbin[1:0] == 'h2);
			Dfourth <= #1 (rbin[1:0] == 'h3);
			ram_read <= #1 enable_ram_read;
			if (enable_state) state <= #1 ENABLE;
			else state <= #1 WAIT;
			if (enable_ram_read) begin
			  case (rbin[1:0])
			    'h0: begin
				 		latch_value <= latch_value_1;
						value <= ram_read_value_1;
						end
			    'h1: begin
				 		latch_value <= latch_value_2;
						value <= ram_read_value_2;
						end
			    'h2: begin
				 		latch_value <= latch_value_3;
						value <= ram_read_value_3;
						end
			    'h3: begin
				 		latch_value <= latch_value_4;
						value <= ram_read_value_4;
						end
			  endcase
			end
			end
		default: begin
			state <= #1 WAIT;
			Dfirst <= #1 Dfirst;
			Dsecond <= #1 Dsecond;
			Dthird <= #1 Dthird;
			Dfourth <= #1 Dfourth;
			end
	endcase
  end
//
// delay ram_read by 1 clock tick so that it doesn't go away too early
// and we end up writing on a read cycle (in the ramstate FSM)
//
//reg ramread;
//always @ (posedge clock) ramread <= #1 ram_read;
//
// instantiate the 4 rams
//
//wire[1:0] rstate1;
ramstate ram1 (.clock(clock),.nclock(nclock),.resetn(resetn),.enable(Dfirst),.ram_enable(ren),
	.ram_zero(zero),.ram_read(ram_read),.addr_in(address_d),.weight_in(Dvalue),.sum(mem_in_1),
	.read_value(ram_read_value_1),.latch_value(latch_value_1),.rstate(rstate1),.ramwrite(rwrite1));

//wire[1:0] rstate2;
ramstate ram2 (.clock(clock),.nclock(nclock),.resetn(resetn),.enable(Dsecond),.ram_enable(ren),
	.ram_zero(zero),.ram_read(ram_read),.addr_in(address_d),.weight_in(Dvalue),.sum(mem_in_2),
	.read_value(ram_read_value_2),.latch_value(latch_value_2),.rstate(rstate2),.ramwrite(rwrite2));

//wire[1:0] rstate3;
ramstate ram3 (.clock(clock),.nclock(nclock),.resetn(resetn),.enable(Dthird),.ram_enable(ren),
	.ram_zero(zero),.ram_read(ram_read),.addr_in(address_d),.weight_in(Dvalue),.sum(mem_in_3),
	.read_value(ram_read_value_3),.latch_value(latch_value_3),.rstate(rstate3),.ramwrite(rwrite3));

//wire[1:0] rstate4;
ramstate ram4 (.clock(clock),.nclock(nclock),.resetn(resetn),.enable(Dfourth),.ram_enable(ren),
	.ram_zero(zero),.ram_read(ram_read),.addr_in(address_d),.weight_in(Dvalue),.sum(mem_in_4),
	.read_value(ram_read_value_4),.latch_value(latch_value_4),.rstate(rstate4),.ramwrite(rwrite4));
//
// for output...
//
/*
always @ (Dfirst or Dsecond or Dthird or Dfourth or 
	latch_value_1 or latch_value_2 or latch_value_3 or latch_value_4 or 
	ram_read_value_1 or ram_read_value_2 or ram_read_value_3 or ram_read_value_4) begin
  if (Dfirst) begin
     latch_value = latch_value_1;
//	  value = ram_read_value_1;
	  value = mem_in_1;
	  end
  if (Dsecond) begin
     latch_value = latch_value_2;
//	  value = ram_read_value_2;
	  value = mem_in_2;
	  end
  if (Dthird) begin
     latch_value = latch_value_3;
//	  value = ram_read_value_3;
	  value = mem_in_3;
	  end
  if (Dfourth) begin
     latch_value = latch_value_4;
//	  value = ram_read_value_4;
	  value = mem_in_4;
	  end
  end
*/
assign rv1 = ram_read_value_1;
assign rv2 = ram_read_value_2;
assign rv3 = ram_read_value_3;
assign rv4 = ram_read_value_4;
endmodule
